{* rijndaeltest_ref   v2.0   August '99 *}

(* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *                ---------------------------------                *
 *                            DELPHI                               *
 *         Implementation of the Known Answer Tests (KAT)          *
 *                   and Monte Carlo Test (MCT)                    *
 *                ---------------------------------                *
 *                                                   December 2000 *
 *                                                                 *
 * Authors: Paulo Barreto                                          *
 *          Vincent Rijmen                                         *
 *                                                                 *
 * Delphi translation by Sergey Kirichenko (ksv@cheerful.com)      *
 *                                                                 *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *)


program rijndaeltest_ref;
{$APPTYPE CONSOLE}

{$define TRACE_KAT_MCT}

uses
  SysUtils,
  rijndael_api_ref;

const
  SUBMITTER = 'Joan Daemen';

function iif(bExpression: boolean; iResTrue,iResFalse: integer): integer;
begin
  if bExpression then
    result:= iResTrue
  else
    result:= iResFalse;
end;

function iif2(bExpression: boolean; sResTrue, sResFalse: string): string;
begin
  if bExpression then
    result:= sResTrue
  else
    result:= sResFalse;
end;

procedure wait;
begin
  write('* hit <enter> to exit...'); readln; writeln;
end;

procedure haltwait;
begin
  wait; halt;
end;

procedure blockPrint(var fp: textfile; const block: pBYTE; blockBits: integer; tag: string);
var
  i: integer;
begin
  write(fp, format('%s=',[tag]));
  for i:= 0 to (blockBits div 8)-1 do
    write(fp, format('%.2X',[block[i]]));
  writeln(fp,'');
  flush(fp);
end; {* blockPrint *}

procedure HexToBin(binBlock: pBYTE; const hexBlock: pchar; blockLen: integer);
var
  i,k,j,t: integer;
begin
  j:= 0;
  i:= 0;
  k:= 0;
  while (blockLen > 0) do
    begin
      t:= integer(hexBlock[i]);
      inc(i);
      if ((t >= ord('0')) and (t <= ord('9'))) then
        j:= (t - ord('0')) shl 4
      else
        if ((t >= ord('a')) and (t <= ord('f'))) then
          j:= (t - ord('a') + 10) shl 4
        else
          if ((t >= ord('A')) and (t <= ord('F'))) then
            j:= (t - ord('A') + 10) shl 4;

      t:= integer(hexBlock[i]);
      inc(i);
      if ((t >= ord('0')) and (t <= ord('9'))) then
        j:= j xor (t - ord('0'))
      else
        if ((t >= ord('a')) and (t <= ord('f'))) then
          j:= j xor (t - ord('a') + 10)
        else
          if ((t >= ord('A')) and (t <= ord('F'))) then
            j:= j xor (t - ord('A') + 10);

    binBlock[k]:= j;
    inc(k);
    blockLen:= blockLen- 8;
  end;
end; {* HexToBin *}

procedure rijndaelVKKAT(var fp: textfile; keyLength, blockLength: integer);
var
  i, j, r: integer;
  block: array [0..(4*MAXBC)-1] of BYTE;
  keyMaterial: array [0..320-1] of BYTE;
  byteVal: BYTE;
  keyInst: TkeyInstance;
  cipherInst: TcipherInstance;
begin
  byteVal:= BYTE('8');
{$ifdef TRACE_KAT_MCT}
  write(format('Executing Variable-Key KAT (key %d): ',[keyLength]));
{$endif} // ?TRACE_KAT_MCT
  write(fp,format(#13+#10+
		'=========='+#13+#10+
		#13+#10+
		'KEYSIZE=%d'+#13+#10+
		#13+#10,[keyLength]));
  flush(fp);
  FillChar(block, blockLength div 8, 0); // memset
  blockPrint(fp, addr(block), blockLength, 'PT');
  FillChar(keyMaterial, sizeof(keyMaterial),0);
  FillChar(keyMaterial, keyLength div 4,'0');
  for i:= 0 to keyLength-1 do
    begin
      keyMaterial[i div 4]:= byteVal; { set only the i-th bit of the i-th test key }
      keyInst.blockLen:= blockLength;
      r:= makeKey(addr(keyInst), DIR_ENCRYPT, keyLength, addr(keyMaterial));
      if (rTRUE <> r)  then
        begin
          writeln(format('makeKey error %d',[r]));
          haltwait; //(-1);
        end;
      writeln(fp, format(#13+#10+'I=%d', [i+1]));
      writeln(fp, format('KEY=%s',[StrPas(addr(keyMaterial))]));
      FillChar(block, blockLength div 8, 0);
      cipherInst.blockLen:= blockLength;
      r:= cipherInit(addr(cipherInst), MODE_ECB, nil);
      if (rTRUE <> r) then
        begin
          writeln(format('cipherInit error %d',[r]));
          haltwait; //(-1);
        end;
      r:= blocksEnCrypt(addr(cipherInst), addr(keyInst), addr(block), blockLength, addr(block));
      if (blockLength <> r) then
        begin
          writeln(format('blockEncrypt error %d',[r]));
          haltwait; //(-1);
        end;
      blockPrint(fp, addr(block), blockLength, 'CT');
      { now check decryption: }
      keyInst.blockLen:= blockLength;
      makeKey(addr(keyInst), DIR_DECRYPT, keyLength, addr(keyMaterial));
      blocksDeCrypt(addr(cipherInst), addr(keyInst), addr(block), blockLength, addr(block));
      for j:= 0 to (blockLength div 8)-1 do
        begin
          assert(block[j] = 0);      {!debug!}
        end;
      { undo changes for the next iteration: }
      keyMaterial[i div 4]:= BYTE('0');
      byteVal:=
        iif(byteVal = ord('8'), ord('4'),
        iif(byteVal = ord('4'), ord('2'),
        iif(byteVal = ord('2'), ord('1'),
        {(byteVal == '1')} ord('8'))));
    end;
  assert(byteVal = BYTE('8'));      {!debug!}
{$ifdef TRACE_KAT_MCT}
  writeln(' done.');
{$endif} // ?TRACE_KAT_MCT
end; {* rijndaelVKKAT *}

procedure rijndaelVTKAT(var fp: textfile; keyLength, blockLength: integer);
var
  i: integer;
  block: array [0..(4*MAXBC)-1] of BYTE;
  keyMaterial: array [0..320-1] of BYTE;
  keyInst: TkeyInstance;
  cipherInst: TcipherInstance;
begin
{$ifdef TRACE_KAT_MCT}
  write(format('Executing Variable-Text KAT (key %d): ',[keyLength]));
{$endif} // ?TRACE_KAT_MCT
  writeln(fp,format(#13+#10+
		'=========='+#13+#10+
		#13+#10+
		'KEYSIZE=%d'+#13+#10
		,[keyLength]));
  flush(fp);
  FillChar(keyMaterial, sizeof(keyMaterial), 0);
  FillChar(keyMaterial, keyLength div 4, '0');
  keyInst.blockLen:= blockLength;
  makeKey(addr(keyInst), DIR_ENCRYPT, keyLength, addr(keyMaterial));
  writeln(fp,format('KEY=%s',[StrPas(addr(keyMaterial))]));
  for i:= 0 to blockLength-1 do
    begin
      FillChar(block, blockLength div 8, 0);
      block[i div 8]:= block[i div 8] or (1 shl (7 - i mod 8)); { set only the i-th bit of the i-th test block }
      writeln(fp,format(#13+#10+'I=%d',[i+1]));
      blockPrint(fp, addr(block), blockLength, 'PT');
      cipherInst.blockLen:= blockLength;
      cipherInit(addr(cipherInst), MODE_ECB, nil);
      blocksEnCrypt(addr(cipherInst), addr(keyInst), addr(block), blockLength, addr(block));
      blockPrint(fp, addr(block), blockLength, 'CT');
    end;
{$ifdef TRACE_KAT_MCT}
  writeln(' done.');
{$endif} // ?TRACE_KAT_MCT
end;

procedure rijndaelTKAT(var fp: textfile; blockLength, keyLength: integer; var fin: textfile);
var
  i, j: integer;
  s: Cardinal; //unsigned int
  block,block2: array [0..(4*MAXBC)-1] of BYTE;
  keyMaterial: array [0..320-1] of BYTE;
  keyInst: TkeyInstance;
  cipherInst: TcipherInstance;
  ch1,ch2: char;
begin
{$ifdef TRACE_KAT_MCT}
  write(format('Executing Tables KAT (key %d): ',[keyLength]));
{$endif} // ?TRACE_KAT_MCT
  writeln(fp,format(#13+#10+
		'=========='+#13+#10+
		#13+#10+
		'KEYSIZE=%d'+
                #13+#10
		,[keyLength]));
  flush(fp);

  FillChar(keyMaterial, sizeof(keyMaterial), 0);
  for i:= 0 to 64-1 do
    begin
      writeln(fp,format(#13+#10+'I=%d', [i+1]));
      for j:= 0 to (keyLength div 4)-1 do
        read(fin,char(keyMaterial[j]));
      keyInst.blockLen:= blockLength;
      makeKey(addr(keyInst), DIR_ENCRYPT, keyLength, addr(keyMaterial));

      writeln(fp,format('KEY=%s',[StrPas(addr(keyMaterial))]));
      for j:= 0 to (blockLength div 8)-1 do
        begin
          read(fin,ch1);
          read(fin,ch1);
          read(fin,ch2);
          s:= strtoint('$'+ch1+ch2);
          block[j]:= s;
        end;
      readln(fin);
      blockPrint(fp, addr(block), blockLength, 'PT');
      cipherInst.blockLen:= blockLength;
      cipherInit(addr(cipherInst), MODE_ECB, nil);
      blocksEnCrypt(addr(cipherInst), addr(keyInst), addr(block), blockLength, addr(block2));
      blockPrint(fp, addr(block2), blockLength, 'CT');
    end;
  for i:= 64 to 128-1 do
    begin
      writeln(fp, format(#13+#10+'I=%d',[i+1]));
      for j:= 0 to (keyLength div 4)-1 do
        read(fin,char(keyMaterial[j]));
      keyInst.blockLen:= blockLength;
      makeKey(addr(keyInst), DIR_DECRYPT, keyLength, addr(keyMaterial));

      writeln(fp,format('KEY=%s',[StrPas(addr(keyMaterial))]));
      for j:= 0 to (blockLength div 8)-1 do
        begin
          read(fin,ch2);
          read(fin,ch1);
          read(fin,ch2);
          s:= strtoint('$'+ch1+ch2);
          block[j]:= s;
        end;
      readln(fin);
      cipherInst.blockLen:= blockLength;
      cipherInit(addr(cipherInst), MODE_ECB, nil);
      blocksDeCrypt(addr(cipherInst), addr(keyInst), addr(block), blockLength, addr(block2));
      blockPrint(fp, addr(block2), blockLength, 'PT');
      blockPrint(fp, addr(block), blockLength, 'CT');
    end;

{$ifdef TRACE_KAT_MCT}
  writeln(' done.');
{$endif} // ?TRACE_KAT_MCT
end;

procedure rijndaelIVKAT(var fp: textfile; keyLength, blockLength: integer);
var
  i, ROUNDS: integer;
  block,block2: array [0..(4*MAXBC)-1] of BYTE;
  keyMaterial: array [0..320-1] of BYTE;
  keyInst: TkeyInstance;
  cipherInst: TcipherInstance;
begin

{$ifdef TRACE_KAT_MCT}
  write(format('Executing Intermediate value KAT (key %d): ',[keyLength]));
{$endif} // ?TRACE_KAT_MCT

  case iif(keyLength >= blockLength, keyLength, blockLength) of
    128: ROUNDS:= 10;
    192: ROUNDS:= 12;
    256: ROUNDS:= 14;
    else
      exit; { this cannot happen }
  end;

  writeln(fp,format(#13+#10+
		'=========='+#13+#10+
		#13+#10+
		'KEYSIZE=%d',
		[keyLength]));
  flush(fp);
  FillChar(keyMaterial, sizeof(keyMaterial), 0);
  for i:= 0 to (keyLength div 8)-1 do
    StrLCopy(addr(keyMaterial[2*i]),pchar(format('%.2X',[i])),2);
  keyInst.blockLen:= blockLength;
  makeKey(addr(keyInst), DIR_ENCRYPT, keyLength, addr(keyMaterial));
  writeln(fp, format('KEY=%s', [StrPas(addr(keyMaterial))]));
  writeln(fp, #13+#10+'Intermediate Ciphertext Values (Encryption)'+#13+#10);
  for i:= 0 to (blockLength div 8)-1 do
    block[i]:= i;
  blockPrint(fp, addr(block), blockLength, 'PT');
  cipherInst.blockLen:= blockLength;
  cipherInit(addr(cipherInst), MODE_ECB, nil);
  for i:= 1 to ROUNDS-1 do
    begin
      cipherUpdateRounds(addr(cipherInst), addr(keyInst), addr(block), blockLength div 8, addr(block2), i);
      blockPrint(fp, addr(block2), blockLength, format('CT%d',[i]));
    end;
  cipherUpdateRounds(addr(cipherInst), addr(keyInst), addr(block), blockLength, addr(block2), ROUNDS);
  blockPrint(fp, addr(block2), blockLength, 'CT');

  keyInst.blockLen:= blockLength;
  makeKey(addr(keyInst), DIR_DECRYPT, keyLength, addr(keyMaterial));
  writeln(fp, #13+#10+'Intermediate Ciphertext Values (Decryption)'+#13+#10);
  blockPrint(fp, addr(block2), blockLength, 'CT');
  cipherInst.blockLen:= blockLength;
  cipherInit(addr(cipherInst), MODE_ECB, nil);
  for i:= 1 to ROUNDS-1 do
    begin
      cipherUpdateRounds(addr(cipherInst), addr(keyInst), addr(block2), blockLength, addr(block), ROUNDS-i);
      blockPrint(fp, addr(block), blockLength, format('PT%d',[i]));
    end;
  cipherUpdateRounds(addr(cipherInst), addr(keyInst), addr(block2), blockLength, addr(block), 0);
  blockPrint(fp, addr(block), blockLength, 'PT');

{$ifdef TRACE_KAT_MCT}
  writeln(' done.');
{$endif}
end;

procedure makeKATs(const vkFile, vtFile, tblFile, ivFile: string);
var
  fp, fp2: textfile;
begin
  { prepare Variable Key Known Answer Tests: }
  AssignFile(fp,vkFile);
  rewrite(fp);           // write mode
  writeln(fp,format(#13+#10+
		'========================='+#13+#10+
		#13+#10+
		'FILENAME:  "%s"'+#13+#10+
		#13+#10+
		'Electronic Codebook (ECB) Mode'+#13+#10+
		'Variable Key Known Answer Tests'+#13+#10+
		#13+#10+
		'Algorithm Name: Rijndael'+#13+#10+
		'Principal Submitter: %s',
		[vkFile,SUBMITTER]));
  flush(fp);

  rijndaelVKKAT(fp, 128, BITSPERBLOCK); { test for 128-bit key }
  rijndaelVKKAT(fp, 192, BITSPERBLOCK); { test for 192-bit key }
  rijndaelVKKAT(fp, 256, BITSPERBLOCK); { test for 256-bit key }

  write(fp,#13+#10+'==========');
  CloseFile(fp);

  { prepare Variable Text Known Answer Tests: }
  AssignFile(fp,vtFile);
  rewrite(fp);           // write mode
  writeln(fp,format(#13+#10+
		'========================='+#13+#10+
		#13+#10+
		'FILENAME:  "%s"'+#13+#10+
		#13+#10+
		'Electronic Codebook (ECB) Mode'+#13+#10+
		'Variable Text Known Answer Tests'+#13+#10+
		#13+#10+
		'Algorithm Name: Rijndael'+#13+#10+
		'Principal Submitter: %s',
		[vtFile,SUBMITTER]));
  flush(fp);

  rijndaelVTKAT(fp, 128, BITSPERBLOCK);
  rijndaelVTKAT(fp, 192, BITSPERBLOCK);
  rijndaelVTKAT(fp, 256, BITSPERBLOCK);

  write(fp,#13+#10+'==========');
  CloseFile(fp);

  { prepare Tables Known Answer Tests: }
  AssignFile(fp,tblFile);
  rewrite(fp);           // write mode
  writeln(fp,format(
	  '/* Description of what tables are tested:'+#13+#10+
          '   The provided implementations each use a different set of tables'+#13+#10+
          '    - Java implementation: uses no tables'+#13+#10+
          '    - reference C implementation: uses Logtable, Alogtable, S, Si, rcon'+#13+#10+
          '    - fast C implementation: uses Logtable, Alogtable,  rcon'+#13+#10+
          '        and additionally, T1, T2, T3, T4, T5, T6, T7, T8'+#13+#10+
          '        and (for the inverse key schedule only) U1, U2, U3, U4.'+#13+#10+
          '   All these tables are tested.'+#13+#10+
          #13+#10+
          '========================='+#13+#10+
          #13+#10+
          'FILENAME:  "%s"'+#13+#10+
          #13+#10+
          'Electronic Codebook (ECB) Mode'+#13+#10+
          'Tables Known Answer Tests'+#13+#10+
          #13+#10+
          'Algorithm Name: Rijndael'+#13+#10+
          'Principal Submitter: %s',
          [tblFile,SUBMITTER]));
  flush(fp);

  if fileexists('table.128') then
    begin
      AssignFile(fp2,'table.128');
      reset(fp2);           // read mode
      rijndaelTKAT(fp, 128, 128, fp2);
      CloseFile(fp2);
    end
  else
    begin
      writeln('Table Known Answer test expects file table.128');
      haltwait; // (EXIT_FAILURE);
    end;
  if fileexists('table.192') then
    begin
      AssignFile(fp2,'table.192');
      reset(fp2);           // read mode
      rijndaelTKAT(fp, 128, 192, fp2);
      CloseFile(fp2);
    end
  else
    begin
      writeln('Table Known Answer test expects file table.192');
      haltwait; //(EXIT_FAILURE);
    end;

  if fileexists('table.256') then
    begin
      AssignFile(fp2,'table.256');
      reset(fp2);           // read mode
      rijndaelTKAT(fp, 128, 256, fp2);
      CloseFile(fp2);
    end
  else
    begin
      writeln('Table Known Answer test expects file table.256');
      haltwait; // (EXIT_FAILURE);
    end;

  write(fp,#13+#10+'==========');
  CloseFile(fp);

  { prepare Intermediate Values Known Answer Tests: }
  AssignFile(fp,ivFile);
  rewrite(fp);
  writeln(fp,format(#13+#10+
		'========================='+#13+#10+
		#13+#10+
		'FILENAME:  "%s"'+#13+#10+
		#13+#10+
		'Electronic Codebook (ECB) Mode'+#13+#10+
		'Intermediate Value Known Answer Tests'+#13+#10+
		#13+#10+
		'Algorithm Name: Rijndael'+#13+#10+
		'Principal Submitter: %s',
		[ivFile,SUBMITTER]));
  flush(fp);

  rijndaelIVKAT(fp, 128, BITSPERBLOCK);
  rijndaelIVKAT(fp, 192, BITSPERBLOCK);
  rijndaelIVKAT(fp, 256, BITSPERBLOCK);

  write(fp,#13+#10+'==========');
  CloseFile(fp);
end;

procedure rijndaelECB_MCT(var fp: textfile; const initKey: pchar; keyLength: integer;
	const initBlock: pchar; blockLength: integer; direction: byte);
var
  i, j: integer;
  inBlock, outBlock, binKey: array [0..(4*MAXBC)-1] of BYTE;
  keyMaterial: array [0..320-1] of byte;
  keyInst: TkeyInstance;
  cipherInst: TcipherInstance;
{$ifdef TRACE_KAT_MCT}
  elapsed: TDateTime;
{$endif} // ?TRACE_KAT_MCT
begin

{$ifdef TRACE_KAT_MCT}
  write(format('Executing ECB MCT (%s, key %d): ',[iif2(direction= DIR_ENCRYPT,'ENCRYPT','DECRYPT'),keyLength]));
  elapsed:= time;
{$endif} // ?TRACE_KAT_MCT
  writeln(fp,format(#13+#10+
		'========================='+#13+#10+
		#13+#10+
		'KEYSIZE=%d',[keyLength]));
  flush(fp);
  HexToBin(addr(outBlock), initBlock, blockLength);
  HexToBin(addr(binKey), initKey, keyLength);
  for i:= 0 to 400-1 do
    begin
{$ifdef TRACE_KAT_MCT}
      write('.');
{$endif} // ?TRACE_KAT_MCT
      writeln(fp, format(#13+#10+'I=%d',[i]));
      { prepare key: }
      for j:= 0 to (keyLength div 8)-1 do
        StrLCopy(addr(keyMaterial[2*j]),pchar(format('%.2X',[binKey[j]])),2);
      keyMaterial[keyLength div 4]:= 0;
      writeln(fp, format('KEY=%s', [StrPas(addr(keyMaterial))]));
      keyInst.blockLen:= blockLength;
      makeKey(addr(keyInst), direction, keyLength, addr(keyMaterial));
      { do encryption/decryption: }
      blockPrint(fp, addr(outBlock), blockLength, iif2(direction = DIR_ENCRYPT, 'PT', 'CT'));
      cipherInst.blockLen:= blockLength;
      cipherInit(addr(cipherInst), MODE_ECB, nil);
      if (direction = DIR_ENCRYPT) then
        for j:= 0 to 10000-1 do
          begin
            Move(outBlock, inBlock, blockLength div 8); //memcpy
            blocksEnCrypt(addr(cipherInst), addr(keyInst), addr(inBlock), blockLength, addr(outBlock));
          end
      else
        for j:= 0 to 10000-1 do
          begin
            Move(outBlock, inBlock, blockLength div 8);  //memcpy
            blocksDeCrypt(addr(cipherInst), addr(keyInst), addr(inBlock), blockLength, addr(outBlock));
          end;
      blockPrint(fp, addr(outBlock), blockLength, iif2(direction = DIR_ENCRYPT, 'CT', 'PT'));
      { prepare new key: }
      case keyLength of
        128: for j:= 0 to (128 div 8)-1 do
               binKey[j]:= binKey[j] xor outBlock[j];
        192: begin
               for j:= 0 to (64 div 8)-1 do
                 binKey[j]:= binKey[j] xor inBlock[j + (64 div 8)];
               for j:= 0 to (128 div 8)-1 do
                 binKey[j + (64 div 8)]:= binKey[j + (64 div 8)] xor outBlock[j];
             end;
        256: begin
               for j:= 0 to (128 div 8)-1 do
                 binKey[j]:= binKey[j] xor inBlock[j];
               for j:= 0 to (128 div 8)-1 do
                 binKey[j + (128 div 8)]:= binKey[j + (128 div 8)] xor outBlock[j];
             end;
      end;
    end;
{$ifdef TRACE_KAT_MCT}
  elapsed:= time -elapsed;
  writeln(format(' done (%s).', [timetostr(elapsed)]));
{$endif} // ?TRACE_KAT_MCT
end; {* rijndaelECB_MCT *}

procedure rijndaelCBC_MCT(var fp: textfile; const initKey: pchar; keyLength: integer;
	const initIV, initBlock: pchar; blockLength: integer; direction: BYTE);
var
  i, j, r, t: integer;
  inBlock, outBlock, binKey, cv: array [0..(256 div 8)-1] of byte;
  keyMaterial: array [0..320-1] of byte;
  iv: array [0..64] of byte;
  keyInst: TkeyInstance;
  cipherInst: TcipherInstance;
{$ifdef TRACE_KAT_MCT}
  elapsed: TDateTime;
{$endif} // ?TRACE_KAT_MCT
begin

{$ifdef TRACE_KAT_MCT}
  write(format('Executing CBC MCT (%s, key %d): ',[iif2(direction= DIR_ENCRYPT,'ENCRYPT','DECRYPT'),keyLength]));
  elapsed:= time;
{$endif} // ?TRACE_KAT_MCT
  writeln(fp,format(#13+#10+
		'=========='+#13+#10+
		#13+#10+
		'KEYSIZE=%d',[keyLength]));
  flush(fp);
  HexToBin(addr(inBlock), initBlock, blockLength); { this is either PT0 or CT0 }
  HexToBin(addr(cv), initIV, blockLength);
  HexToBin(addr(binKey), initKey, keyLength);
  for i:= 0 to 400-1 do
    begin
{$ifdef TRACE_KAT_MCT}
      write('.');
{$endif} // ?TRACE_KAT_MCT
      writeln(fp, format(#13+#10+'I=%d',[i]));
      { prepare key: }
      for j:= 0 to (keyLength div 8)-1 do
        StrLCopy(addr(keyMaterial[2*j]),pchar(format('%.2X',[binKey[j]])),2);
      keyMaterial[keyLength div 4]:= 0;
      writeln(fp, format('KEY=%s', [StrPas(addr(keyMaterial))]));
      keyInst.blockLen:= blockLength;
      r:= makeKey(addr(keyInst), direction, keyLength, addr(keyMaterial));
      if (rTRUE <> r) then
        begin
          writeln(format('makeKey error %d',[r]));
          haltwait; //-1
        end;
      { do encryption/decryption: }
      blockPrint(fp, addr(cv), blockLength, 'IV');
      blockPrint(fp, addr(inBlock), blockLength, iif2(direction = DIR_ENCRYPT, 'PT', 'CT'));
      if (direction = DIR_ENCRYPT) then
        for j:= 0 to 10000-1 do
          begin
            for t:= 0 to (blockLength div 8)-1 do
              StrLCopy(addr(iv[2*t]),pchar(format('%.2X',[cv[t]])),2);
            cipherInst.blockLen:= blockLength;
            r:= cipherInit(addr(cipherInst), MODE_CBC, addr(iv));
            if (rTRUE <> r) then
              begin
                writeln(format('cipherInit error %d',[r]));
                haltwait;
              end;
            r:= blocksEnCrypt(addr(cipherInst), addr(keyInst), addr(inBlock), blockLength, addr(outBlock));
            if (blockLength <> r) then
              begin
                writeln(format('blockEncrypt error %d',[r]));
                haltwait;
              end;
            Move(cv, inBlock, blockLength div 8);
            Move(outBlock, cv, blockLength div 8);
          end
      else
        for j:= 0 to 10000-1 do
          begin
            for t:= 0 to (blockLength div 8)-1 do
              StrLCopy(addr(iv[2*t]),pchar(format('%.2X',[cv[t]])),2);
            cipherInst.blockLen:= blockLength;
            cipherInit(addr(cipherInst), MODE_CBC, addr(iv));
            blocksDeCrypt(addr(cipherInst), addr(keyInst), addr(inBlock), blockLength, addr(outBlock));
            Move(inBlock, cv, blockLength div 8);
            Move(outBlock, inBlock, blockLength div 8);
          end;
      blockPrint(fp, addr(outBlock), blockLength, iif2(direction = DIR_ENCRYPT, 'CT', 'PT'));
      { prepare new key: }
      case (keyLength) of
        128:  for j:= 0 to (128 div 8)-1 do
                binKey[j]:= binKey[j] xor outBlock[j];
        192:  begin
                for j:= 0 to (64 div 8)-1 do
                  if (direction = DIR_ENCRYPT) then
                    binKey[j]:= binKey[j] xor inBlock[j + (64 div 8)]
                  else
                    binKey[j]:= binKey[j] xor cv[j + (64 div 8)];
                for j:= 0 to (128 div 8)-1 do
                  binKey[j+ (64 div 8)]:= binKey[j+ (64 div 8)] xor outBlock[j];
              end;
        256:  begin
                for j:= 0 to (128 div 8)-1 do
                  if (direction = DIR_ENCRYPT) then
                    binKey[j]:= binKey[j] xor inBlock[j]
                  else
                    binKey[j]:= binKey[j] xor cv[j];
                for j:= 0 to (128 div 8)-1 do
                  binKey[j+ (128 div 8)]:= binKey[j+ (128 div 8)] xor outBlock[j];
              end;
      end;
    end;
{$ifdef TRACE_KAT_MCT}
  elapsed:= time -elapsed;
  writeln(format(' done (%s).', [timetostr(elapsed)]));
{$endif} // ?TRACE_KAT_MCT
end; {* rijndaelCBC_MCT *}


procedure makeMCTs(const ecbEncryptionFile, ecbDecryptionFile,
	cbcEncryptionFile, cbcDecryptionFile: string);
var
  fp: textfile;
begin
  { prepare ECB Encryption Monte Carlo Tests: }
  AssignFile(fp,ecbEncryptionFile);
  rewrite(fp);
  writeln(fp,format(#13+#10+
		'========================='+#13+#10+
		#13+#10+
		'FILENAME:  "%s"'+#13+#10+
		#13+#10+
		'Electronic Codebook (ECB) Mode - ENCRYPTION'+#13+#10+
		'Monte Carlo Test'+#13+#10+
		#13+#10+
		'Algorithm Name: Rijndael'+#13+#10+
		'Principal Submitter: %s',
		[ecbEncryptionFile,SUBMITTER]));
  flush(fp);
  rijndaelECB_MCT(fp,
		'00000000000000000000000000000000', 128,
		'00000000000000000000000000000000', BITSPERBLOCK, DIR_ENCRYPT);
  rijndaelECB_MCT(fp,
		'000000000000000000000000000000000000000000000000', 192,
		'00000000000000000000000000000000', BITSPERBLOCK, DIR_ENCRYPT);
  rijndaelECB_MCT(fp,
		'0000000000000000000000000000000000000000000000000000000000000000', 256,
		'00000000000000000000000000000000', BITSPERBLOCK, DIR_ENCRYPT);
  write(fp,#13+#10+'===========');
  CloseFile(fp);

  { prepare ECB Decryption Monte Carlo Tests: }
  AssignFile(fp,ecbDecryptionFile);
  rewrite(fp);
  writeln(fp,format(#13+#10+
		'========================='+#13+#10+
		#13+#10+
		'FILENAME:  "%s"'+#13+#10+
		#13+#10+
		'Electronic Codebook (ECB) Mode - DECRYPTION'+#13+#10+
		'Monte Carlo Test'+#13+#10+
		#13+#10+
		'Algorithm Name: Rijndael'+#13+#10+
		'Principal Submitter: %s',
		[ecbDecryptionFile,SUBMITTER]));
  flush(fp);
  rijndaelECB_MCT(fp,
		'00000000000000000000000000000000', 128,
		'00000000000000000000000000000000', BITSPERBLOCK, DIR_DECRYPT);
  rijndaelECB_MCT(fp,
		'000000000000000000000000000000000000000000000000', 192,
		'00000000000000000000000000000000', BITSPERBLOCK, DIR_DECRYPT);
  rijndaelECB_MCT(fp,
		'0000000000000000000000000000000000000000000000000000000000000000', 256,
		'00000000000000000000000000000000', BITSPERBLOCK, DIR_DECRYPT);
  write(fp,#13+#10+'===========');
  CloseFile(fp);

  { prepare CBC Encryption Monte Carlo Tests: }
  AssignFile(fp,cbcEncryptionFile);
  rewrite(fp);
  writeln(fp,format(#13+#10+
  		'========================='+#13+#10+
		#13+#10+
		'FILENAME:  "%s"'+#13+#10+
		#13+#10+
		'Cipher Block Chaining (CBC) Mode - ENCRYPTION'+#13+#10+
		'Monte Carlo Test'+#13+#10+
		#13+#10+
		'Algorithm Name: Rijndael'+#13+#10+
		'Principal Submitter: %s',
		[cbcEncryptionFile,SUBMITTER]));
  flush(fp);
  rijndaelCBC_MCT(fp,
		'00000000000000000000000000000000', 128,
		'00000000000000000000000000000000', '00000000000000000000000000000000', BITSPERBLOCK,
		DIR_ENCRYPT);
  rijndaelCBC_MCT(fp,
		'000000000000000000000000000000000000000000000000', 192,
		'00000000000000000000000000000000', '00000000000000000000000000000000', BITSPERBLOCK,
		DIR_ENCRYPT);
  rijndaelCBC_MCT(fp,
		'0000000000000000000000000000000000000000000000000000000000000000', 256,
		'00000000000000000000000000000000', '00000000000000000000000000000000', BITSPERBLOCK,
		DIR_ENCRYPT);
  write(fp,#13+#10+'===========');
  CloseFile(fp);

  { prepare CBC Decryption Monte Carlo Tests: }
  AssignFile(fp,cbcDecryptionFile);
  rewrite(fp);
  writeln(fp,format(#13+#10+
		'========================='+#13+#10+
		#13+#10+
		'FILENAME:  "%s"'+#13+#10+
		#13+#10+
		'Cipher Block Chaining (CBC) Mode - DECRYPTION'+#13+#10+
		'Monte Carlo Test'+#13+#10+
		#13+#10+
		'Algorithm Name: Rijndael'+#13+#10+
		'Principal Submitter: %s',
		[cbcDecryptionFile,SUBMITTER]));
  flush(fp);
  rijndaelCBC_MCT(fp,
		'00000000000000000000000000000000', 128,
		'00000000000000000000000000000000', '00000000000000000000000000000000', BITSPERBLOCK,
		DIR_DECRYPT);
  rijndaelCBC_MCT(fp,
		'000000000000000000000000000000000000000000000000', 192,
		'00000000000000000000000000000000', '00000000000000000000000000000000', BITSPERBLOCK,
		DIR_DECRYPT);
  rijndaelCBC_MCT(fp,
		'0000000000000000000000000000000000000000000000000000000000000000', 256,
		'00000000000000000000000000000000', '00000000000000000000000000000000', BITSPERBLOCK,
		DIR_DECRYPT);
  write(fp,#13+#10+'===========');
  CloseFile(fp);
end; {* makeMCTs *}

begin
{* main *}
  makeKATs('ecb_vk.txt', 'ecb_vt.txt', 'ecb_tbl.txt', 'ecb_iv.txt');
  makeMCTs('ecb_e_m.txt', 'ecb_d_m.txt', 'cbc_e_m.txt', 'cbc_d_m.txt');
  wait;
end.
